import torch
import numpy as np
torch.manual_seed(0)
from torch import nn
from torch.autograd import Function
from torch.autograd import Variable
from NURBSDiff.surf_eval_cpp import pre_compute_basis as cpp_pre_compute_basis, forward as cpp_forward, backward as cpp_backward
CUDA_AVAILABLE=True
DELTA = 1e-8
try:
    from NURBSDiff.surf_eval_cuda import pre_compute_basis, forward, backward
except:
    CUDA_AVAILABLE=False
from .utils import gen_knot_vector

class SurfEval(torch.nn.Module):
    """
    We can implement our own custom autograd Functions by subclassing
    torch.autograd.Function and implementing the forward and backward passes
    which operate on Tensors.
    """
    def __init__(self, m, n, dimension=3, p=3, q=3, knot_u=None, knot_v=None, out_dim_u=32, out_dim_v=128, method='tc', dvc='cpp'):
        super(SurfEval, self).__init__()
        if not CUDA_AVAILABLE:
            dvc = 'cpp'
        self.m = m
        self.n = n
        self._dimension = dimension
        self.p, self.q = p, q
        if knot_u is not None:
            self.U = torch.Tensor(knot_u)
        else:
            self.U = torch.Tensor(np.array(gen_knot_vector(self.p, self.m)))
        if knot_v is not None:
            self.V = torch.Tensor(knot_v)
        else:
            self.V = torch.Tensor(np.array(gen_knot_vector(self.q, self.n)))
        self.u = torch.linspace(0.0 + DELTA, 1.0 - DELTA, steps=out_dim_u, dtype=torch.float32)
        self.v = torch.linspace(0.0 + DELTA, 1.0 - DELTA, steps=out_dim_v, dtype=torch.float32)
        self.method = method
        self.dvc = dvc
        if self.dvc == 'cuda':
            self.U = self.U.cuda()
            self.u = self.u.cuda()
            self.V = self.V.cuda()
            self.v = self.v.cuda()
            self.uspan_uv, self.vspan_uv, self.Nu_uv, self.Nv_uv = pre_compute_basis(self.u, self.v, self.U, self.V, m, n, p , q, out_dim_u, self._dimension)
            self.Nu_uv = self.Nu_uv.view(out_dim_u, p+1)
            self.Nv_uv = self.Nv_uv.view(out_dim_v, q+1)
        else:
            self.uspan_uv, self.vspan_uv, self.Nu_uv, self.Nv_uv = cpp_pre_compute_basis(self.u, self.v, self.U, self. V, m, n, p , q, out_dim_u, self._dimension)
            self.Nu_uv = self.Nu_uv.view(out_dim_u, p+1)
            self.Nv_uv = self.Nv_uv.view(out_dim_v, q+1)

        # if self.method == 'tc':
        #     self.Nu_uv = self.Nu_uv.repeat(self.v.size(0), 1, 1)
        #     self.Nv_uv = self.Nv_uv.repeat(self.u.size(0), 1, 1)

    def forward(self,input):
        """
        In the forward pass we receive a Tensor containing the input and return
        a Tensor containing the output. ctx is a context object that can be used
        to stash information for backward computation. You can cache arbitrary
        objects for use in the backward pass using the ctx.save_for_backward method.
        """
        # input will be of dimension (batch_size, m+1, n+1, dimension)

        if self.method == 'cpp':
            out = SurfEvalFunc.apply(input, self.uspan_uv, self.vspan_uv, self.Nu_uv, self.Nv_uv, self.u, self.v, self.m, self.n, self.p, self.q, self._dimension, self.dvc)
            return out
        elif self.method == 'tc':
            surfaces = (self.Nu_uv[:,0].unsqueeze(0).unsqueeze(-1).unsqueeze(-1)*\
                input[:,(self.uspan_uv - self.p).type(torch.LongTensor), :,:])[:,:, (self.vspan_uv-self.q).type(torch.LongTensor),:]*\
                self.Nv_uv[:,0].unsqueeze(0).unsqueeze(0).unsqueeze(-1)

            for r in range(1,self.q+1):
                surfaces += (self.Nu_uv[:,0].unsqueeze(0).unsqueeze(-1).unsqueeze(-1)*\
                    input[:,(self.uspan_uv - self.p).type(torch.LongTensor), :,:])[:,:, (self.vspan_uv-self.q+r).type(torch.LongTensor),:]*\
                    self.Nv_uv[:,r].unsqueeze(0).unsqueeze(0).unsqueeze(-1)

            for l in range(1,self.p+1):
                for r in range(self.q+1):
                    surfaces += (self.Nu_uv[:,l].unsqueeze(0).unsqueeze(-1).unsqueeze(-1)*\
                        input[:,(self.uspan_uv - self.p+l).type(torch.LongTensor), :,:])[:,:, (self.vspan_uv-self.q+r).type(torch.LongTensor),:]*\
                        self.Nv_uv[:,r].unsqueeze(0).unsqueeze(0).unsqueeze(-1)

            surfaces = surfaces[:,:,:,:self._dimension]/surfaces[:,:,:,self._dimension].unsqueeze(-1)
            return surfaces



class SurfEvalFunc(torch.autograd.Function):

    @staticmethod
    def forward(ctx, ctrl_pts, uspan_uv, vspan_uv, Nu_uv, Nv_uv, u_uv, v_uv, m, n, p, q, _dimension, _device):
        ctx.save_for_backward(ctrl_pts)
        ctx.uspan_uv = uspan_uv
        ctx.vspan_uv = vspan_uv
        ctx.Nu_uv = Nu_uv
        ctx.Nv_uv = Nv_uv
        ctx.u_uv = u_uv
        ctx.v_uv = v_uv
        ctx.m = m
        ctx.n = n
        ctx.p = p
        ctx.q = q
        ctx._dimension = _dimension
        ctx._device = _device

        if _device == 'cuda':
            surfaces = forward(ctrl_pts, uspan_uv, vspan_uv, Nu_uv, Nv_uv, u_uv, v_uv, m, n, p, q, _dimension)
        else:
            surfaces = cpp_forward(ctrl_pts, uspan_uv, vspan_uv, Nu_uv, Nv_uv, u_uv, v_uv, m, n, p, q, _dimension)

        ctx.surfaces=surfaces
        return surfaces[:,:,:,:_dimension]/surfaces[:,:,:,_dimension].unsqueeze(-1)

    @staticmethod
    def backward(ctx, grad_output):
        ctrl_pts,  = ctx.saved_tensors
        uspan_uv = ctx.uspan_uv
        vspan_uv = ctx.vspan_uv
        Nu_uv = ctx.Nu_uv
        Nv_uv = ctx.Nv_uv
        u_uv = ctx.u_uv
        v_uv = ctx.v_uv
        m = ctx.m
        n = ctx.n
        p = ctx.p
        q = ctx.q
        _dimension = ctx._dimension
        _device = ctx._device
        surfaces=ctx.surfaces
        grad_sw = torch.zeros((grad_output.size(0),grad_output.size(1),grad_output.size(2),_dimension+1),dtype=torch.float32)
        grad_sw[:,:,:,:_dimension] = grad_output
        if _device == 'cuda':
            grad_sw = grad_sw.cuda()

        for d in range(_dimension):
            grad_sw[:,:,:,_dimension] += grad_output[:,:,:,d]/surfaces[:,:,:,_dimension]


        if _device == 'cuda':
            grad_ctrl_pts = backward(grad_sw, ctrl_pts, uspan_uv, vspan_uv, Nu_uv, Nv_uv, u_uv, v_uv, m, n, p, q, _dimension)
        else:
            grad_ctrl_pts = cpp_backward(grad_sw, ctrl_pts, uspan_uv, vspan_uv, Nu_uv, Nv_uv, u_uv, v_uv, m, n, p, q, _dimension)

        
        
        return Variable(grad_ctrl_pts[0]), None, None, None, None, None, None,None,None,None,None,None,None,None




import torch
import numpy as np
torch.manual_seed(0)
from torch import nn
from torch.autograd import Function
from torch.autograd import Variable

from .utils import gen_knot_vector

from NURBSDiff.surf_eval_cpp import pre_compute_basis as cpp_pre_compute_basis, forward as cpp_forward, backward as cpp_backward
from NURBSDiff.surf_eval_cuda import pre_compute_basis, forward, backward

class SurfEval(torch.nn.Module):
    """
    We can implement our own custom autograd Functions by subclassing
    torch.autograd.Function and implementing the forward and backward passes
    which operate on Tensors.
    """
    def __init__(self, m, n, dimension=3, p=3, q=3, out_dim_u=32, out_dim_v=128, method='tc', dvc='cpp'):
        super(SurfEval, self).__init__()
        self.m = m
        self.n = n
        self._dimension = dimension
        self.p, self.q = p, q
        self.out_dim_u, self.out_dim_v = out_dim_u, out_dim_v
        self.u = torch.linspace(1e-5, 1.0-1e-5, steps=out_dim_u, dtype=torch.float32)
        self.v = torch.linspace(1e-5, 1.0-1e-5, steps=out_dim_v, dtype=torch.float32)
        self.method = method
        self.dvc = dvc
        if self.dvc == 'cuda':
            self.u = self.u.cuda()
            self.v = self.v.cuda()


    def forward(self,input):
        """
        In the forward pass we receive a Tensor containing the input and return
        a Tensor containing the output. ctx is a context object that can be used
        to stash information for backward computation. You can cache arbitrary
        objects for use in the backward pass using the ctx.save_for_backward method.
        """
        # input will be of dimension (batch_size, m+1, n+1, dimension)
        ctrl_pts, knot_u, knot_v = input

        U_c = torch.cumsum(torch.where(knot_u<0.0, knot_u*0+1e-4, knot_u), dim=1)
        U = (U_c - U_c[:,0].unsqueeze(-1)) / (U_c[:,-1].unsqueeze(-1) - U_c[:,0].unsqueeze(-1))
        V_c = torch.cumsum(torch.where(knot_v<0.0, knot_v*0+1e-4, knot_v), dim=1)

        V = (V_c - V_c[:,0].unsqueeze(-1)) / (V_c[:,-1].unsqueeze(-1) - V_c[:,0].unsqueeze(-1))

        if torch.isnan(V).any():
            print(V_c)
            print(knot_v)

        if torch.isnan(U).any():
            print(U_c)
            print(knot_u)

        #############################################################################
        #################### Gaussian gradient smoothening ##########################

        # u = self.u.unsqueeze(0)
        # v = self.v.unsqueeze(0)

        # uspan_uv = torch.stack([torch.min(torch.where((u - U[s,self.p:-self.p].unsqueeze(1))>1e-8, u - U[s,self.p:-self.p].unsqueeze(1), (u - U[s,self.p:-self.p].unsqueeze(1))*0.0 + 1),0,keepdim=False)[1]+self.p for s in range(U.size(0))])
        # vspan_uv = torch.stack([torch.min(torch.where((v - V[s,self.q:-self.q].unsqueeze(1))>1e-8, v - V[s,self.q:-self.q].unsqueeze(1), (v - V[s,self.q:-self.q].unsqueeze(1))*0.0 + 1),0,keepdim=False)[1]+self.q for s in range(V.size(0))])

        # Nu_uv = BasisFunc.apply(u, U, uspan_uv, self.p)

        # Nu_uv = Nu_uv.unsqueeze(2).unsqueeze(-1).unsqueeze(-1)


        # Nv_uv = BasisFunc.apply(v, V, vspan_uv, self.q)

        # Nv_uv = Nv_uv.unsqueeze(1).unsqueeze(-1).unsqueeze(-3)


        #############################################################################
        #################### Autograd based definition ##############################
        u = self.u.unsqueeze(0)
        uspan_uv = torch.stack([torch.min(torch.where((u - U[s,self.p:-self.p].unsqueeze(1))>1e-8, u - U[s,self.p:-self.p].unsqueeze(1), (u - U[s,self.p:-self.p].unsqueeze(1))*0.0 + 1),0,keepdim=False)[1]+self.p for s in range(U.size(0))])

        u = u.squeeze(0)
        Ni = [u*0 for i in range(self.p+1)]
        Ni[0] = u*0 + 1
        for k in range(1,self.p+1):
            saved = (u)*0.0
            for r in range(k):
                UList1 = torch.stack([U[s,uspan_uv[s,:] + r + 1] for s in range(U.size(0))])
                UList2 = torch.stack([U[s,uspan_uv[s,:] + 1 - k + r] for s in range(U.size(0))])
                temp = Ni[r]/((UList1 - u) + (u - UList2))
                temp = torch.where(((UList1 - u) + (u - UList2))==0.0, u*0+1e-4, temp)
                Ni[r] = saved + (UList1 - u)*temp
                saved = (u - UList2)*temp
            Ni[k] = saved

        Nu_uv = torch.stack(Ni).permute(1,0,2).unsqueeze(2).unsqueeze(-1).unsqueeze(-1)

        v = self.v.unsqueeze(0)
        vspan_uv = torch.stack([torch.min(torch.where((v - V[s,self.q:-self.q].unsqueeze(1))>1e-8, v - V[s,self.q:-self.q].unsqueeze(1), (v - V[s,self.q:-self.q].unsqueeze(1))*0.0 + 1),0,keepdim=False)[1]+self.q for s in range(V.size(0))])


        Ni = [v*0 for i in range(self.q+1)]
        Ni[0] = v*0 + 1
        for k in range(1,self.q+1):
            saved = (v)*0.0
            for r in range(k):
                VList1 = torch.stack([V[s,vspan_uv[s,:] + r + 1] for s in range(V.size(0))])
                VList2 = torch.stack([V[s,vspan_uv[s,:] + 1 - k + r] for s in range(V.size(0))])
                temp = Ni[r]/((VList1 - v) + (v - VList2))
                temp = torch.where(((VList1 - v) + (v - VList2))==0.0, v*0+1e-4, temp)
                Ni[r] = saved + (VList1 - v)*temp
                saved = (v - VList2)*temp
            Ni[k] = saved

        Nv_uv = torch.stack(Ni).permute(1,0,2).unsqueeze(1).unsqueeze(-1).unsqueeze(-3)

        ################################################################################
        ################################################################################

        pts = torch.stack([torch.stack([torch.stack([ctrl_pts[s,(uspan_uv[s,:]-self.p+l),:,:][:,(vspan_uv[s,:]-self.q+r),:] \
            for r in range(self.q+1)]) for l in range(self.p+1)]) for s in range(U.size(0))])


        # rational_pts = pts[:, :, :, :, :, :self._dimension]*pts[:, :, :, :, :, self._dimension:]
        # pts = torch.cat((rational_pts,pts[:, :, :, :, :, self._dimension:]),-1)

        # print((Nu_uv*Nv_uv).size(), pts.size())
        surfaces = torch.sum((Nu_uv*pts)*Nv_uv, (1,2))

        # surfaces = torch.sum((Nu_uv*pts), (1,2))
        # print(surfaces[:,:,:,self._dimension].sum())
        # # print(surfaces.size())
        surfaces = surfaces[:,:,:,:self._dimension]#/surfaces[:,:,:,self._dimension].unsqueeze(-1)
        return surfaces



class BasisFunc(torch.autograd.Function):

    @staticmethod
    def forward(ctx, u, U, uspan_uv, p):
        ctx.save_for_backward(U)
        ctx.uspan_uv = uspan_uv
        ctx.u = u
        ctx.p = p

        u = u.squeeze(0)
        Ni = [u*0 for i in range(p+1)]
        Ni[0] = u*0 + 1
        for k in range(1,p+1):
            saved = (u)*0.0
            for r in range(k):
                UList1 = torch.stack([U[s, uspan_uv[s,:] + r + 1] for s in range(U.size(0))])
                UList2 = torch.stack([U[s, uspan_uv[s,:] + 1 - k + r] for s in range(U.size(0))])
                temp = Ni[r]/((UList1 - u) + (u - UList2))
                temp = torch.where(((UList1 - u) + (u - UList2))==0.0, u*0+1e-8, temp)
                Ni[r] = saved + (UList1 - u)*temp
                saved = (u - UList2)*temp
            Ni[k] = saved

        Nu_uv = torch.stack(Ni).permute(1,0,2)
        ctx.Nu_uv = Nu_uv
        return Nu_uv

    @staticmethod
    def backward(ctx, grad_output):
        U = ctx.saved_tensors[0]
        uspan_uv = ctx.uspan_uv
        p = ctx.p
        Nu_uv = ctx.Nu_uv
        u = ctx.u

        UList = torch.stack([U[s, uspan_uv[s,:]] for s in range(U.size(0))])

        dNi = [grad_output[:,0,:]*0 for i in range(p+1)]
        dNi[0] = grad_output[:,0,:]*0 + 0.5*UList*Nu_uv[:,0,:]

        for k in reversed(range(1,p+1)):
            tempdNi = dNi[k]*grad_output[:,k,:]
            for r in reversed(range(k)):
                UList1 = torch.stack([U[s, uspan_uv[s,:] + r + 1] for s in range(U.size(0))])
                UList2 = torch.stack([U[s, uspan_uv[s,:] + 1 - k + r] for s in range(U.size(0))])
                tempdNi = tempdNi*(u-UList2)
                dNi[r] += (UList1 - u)*tempdNi
                dNi[r] = dNi[r]/((UList1 - u) + (u - UList2))

        dNu_uv = torch.stack(dNi).permute(1,0,2)

        dU = U*0
        for s in range(U.size(0)):
            for k in range(0,p+1):
                dU[s, :].scatter_(-1, (uspan_uv[s,:] + k).type_as(uspan_uv), dNu_uv[s, k, :], reduce='add')
        dU = dU*U


        # for s in range(U.size(0)):
        #     for t in range(uspan_uv.size(1)):
        #         for k in range(1,p+1):
        #             tempdU = dU*0
        #             for r in range(k):
        #                 tempdU[s, uspan_uv[s,t] + r] += grad_output[s, r, t]*U[s, uspan_uv[s,t] + r + 1]
        #                 tempdU[s, uspan_uv[s,:] + r + 1] += (-1)*grad_output[s, 1 - k + r, t]*U[s, uspan_uv[s,:] + 1 - k + r]
        #             dU += tempdU

        # dU = U*0
        # for s in range(U.size(0)):
        #     for k in range(0,p+1):
        #         dU[s, uspan_uv[s,:]] += grad_output[s, k, :]*Nu_uv[s, k, :]*((100/1.00)**2)

        return Variable(U*0), Variable(dU), Variable(U*0), None




import numpy as np


def gen_knot_vector(p,n, delta=1e-6):
    # p: degree, n: number of control points; m+1: number of knots
    m = p + n + 1

    # Calculate a uniform interval for middle knots
    num_segments = (m - 2*(p+1) + 1)  # number of segments in the middle
    spacing = (1.0) / (num_segments)  # spacing between the knots (uniform)

    # First degree+1 knots are "knot_min"
    knot_vector = [(j+1)*delta for j in range(0, p)]

    # Middle knots
    knot_vector += [mid_knot for mid_knot in np.linspace(delta*(p+1), 1, num_segments+1)]

    # Last degree+1 knots are "knot_max"
    knot_vector += [float(1) for j in range(0, p)]

    # Return auto-generated knot vector
    return knot_vector


from setuptools import setup, find_packages
import torch
from torch.utils.cpp_extension import CppExtension, CUDAExtension
import os


class BuildExtension(torch.utils.cpp_extension.BuildExtension):
    def __init__(self, *args, **kwargs):
        super().__init__(use_ninja=False, *args, **kwargs)


os.putenv('TORCH_CUDA_ARCH_LIST', '5.0;5.2;5.3;6.0;6.1;6.2;7.0;7.2;7.5;8.0;8.6;8.7;8.9;9.0')

this_dir = os.path.dirname(os.path.abspath(__file__))
extensions_dir = os.path.join(this_dir, "NURBSDiff", "csrc")
extra_compile_args = {"cxx": ["-std=c++17"]}
include_package = ('NURBSDiff*', 'NURBSDiff/csrc')
try:
    setup(
        ext_modules=[
            CppExtension(name='NURBSDiff.curve_eval_cpp',
                         sources=['NURBSDiff/csrc/curve_eval.cpp', 'NURBSDiff/csrc/utils.cpp'],
                         include_dirs=[extensions_dir],
                         extra_compile_args=extra_compile_args),
            CppExtension(name='NURBSDiff.surf_eval_cpp',
                         sources=['NURBSDiff/csrc/surf_eval.cpp', 'NURBSDiff/csrc/utils.cpp'],
                         include_dirs=[extensions_dir],
                         extra_compile_args=extra_compile_args),
            CUDAExtension(name='NURBSDiff.curve_eval_cuda',
                          sources=['NURBSDiff/csrc/curve_eval_cuda.cpp', 'NURBSDiff/csrc/curve_eval_cuda_kernel.cu'],
                          include_dirs=[extensions_dir],
                          extra_compile_args=extra_compile_args),
            CUDAExtension(name='NURBSDiff.surf_eval_cuda',
                          sources=['NURBSDiff/csrc/surf_eval_cuda.cpp', 'NURBSDiff/csrc/surf_eval_cuda_kernel.cu'],
                          include_dirs=[extensions_dir],
                          extra_compile_args=extra_compile_args),
        ],
        cmdclass={
            'build_ext': BuildExtension
        },
        packages=find_packages(include=include_package, exclude=("examples*", "images*")))
except:
    print('installation of NURBSDiff with GPU wasnt successful, installing CPU version')
    setup(
        ext_modules=[
            CppExtension(name='NURBSDiff.curve_eval_cpp',
                         sources=['NURBSDiff/csrc/curve_eval.cpp', 'NURBSDiff/csrc/utils.cpp'],
                         include_dirs=[extensions_dir],
                         extra_compile_args=extra_compile_args),
            CppExtension(name='NURBSDiff.surf_eval_cpp',
                         sources=['NURBSDiff/csrc/surf_eval.cpp', 'NURBSDiff/csrc/utils.cpp'],
                         include_dirs=[extensions_dir],
                         extra_compile_args=extra_compile_args),
        ],
        cmdclass={
            'build_ext': BuildExtension
        },
        packages=find_packages(include=include_package, exclude=("examples*", "images*")))



import torch
import numpy as np
from torch.autograd import Variable
from NURBSDiff.curve_eval_cpp import forward as cpp_forward, backward as cpp_backward, pre_compute_basis as cpp_pre_compute_basis
from NURBSDiff.curve_eval_cuda import pre_compute_basis, forward, backward
from .utils import gen_knot_vector
torch.manual_seed(120)


class CurveEval(torch.nn.Module):
    """Evaluate control points and weights to sample points on the curvature.
    """

    def __init__(self, num_control_points: int, knot_vectors: torch.Tensor | None = None,
                 dimension: int = 3, order: int = 2, out_dim: int = 32, method: str = 'tc'):
        """Initialize the curve evaluation.

        Args:
            num_control_points (int):
                Number of control points for the curvature.
            knot_vectors (torch.Tensor, optional):
                Optional knot vectors of the curvature. If None are specified,
                it will be automatically generated. Defaults to None.
            dimension (int, optional):
                Dimension of the curvature. Defaults to 3.
            order (int, optional):
                Order of the curvature. Defaults to 2.
            out_dim (int, optional):
                Number of sampled points from the curvature. Defaults to 32.
            method (str, optional):
                Method to use. Choose from 'cpp' or 'tc'. 'cpp' refers to pre-compiled
                c++ code. 'tc' refers to tensor computation in python. Defaults to 'cpp'.
        """
        super(CurveEval, self).__init__()
        self.num_control_points = num_control_points
        self._dimension = dimension
        self.order = order
        if knot_vectors is not None:
            self.knot_vectors = knot_vectors
        else:
            self.knot_vectors = torch.Tensor(self._gen_knot_vector(self.order, self.num_control_points))
        self.curve_param = torch.linspace(0.0, 1.0, steps=out_dim, dtype=torch.float32)
        self.method = method
        self.curve_parameter_span_cpu, self.new_curve_param_cpu = cpp_pre_compute_basis(self.curve_param, self.knot_vectors, num_control_points, order, out_dim, self._dimension)
        self.curve_parameter_span_cuda, self.new_curve_param_cuda = pre_compute_basis(self.curve_param.cuda(), self.knot_vectors.cuda(), num_control_points, order, out_dim, self._dimension)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass of the module.

        Args:
            input (torch.Tensor):
                Control points and weights, shape would be
                (batch_size, num_control_points, control_points_dimension + 1 ),
                where 1 refers to the control point weights.

        Returns:
            torch.Tensor:
                Sampled points from the curvature.
        """
        # input will be of dimension (batch_size, num_control_points+1, n+1, dimension+1)
        device = input.device.type
        if device == 'cuda':
            curve_parameter_span = self.curve_parameter_span_cuda
            new_curve_param = self.new_curve_param_cuda
        else:
            curve_parameter_span = self.curve_parameter_span_cpu
            new_curve_param = self.new_curve_param_cpu

        if self.method == 'cpp':
            if self.curve_param.device.type != device:
                self.curve_param = self.curve_param.to(input.device)
            out = CurveEvalFunc.apply(input, curve_parameter_span, new_curve_param, self.curve_param, self.num_control_points, self.order, self._dimension, device)
            return out
        elif self.method == 'tc':
            # input[:,:,:self._dimension] = input[:,:,:self._dimension]*input[:,:,self._dimension].unsqueeze(-1)
            curves = new_curve_param[:, 0].unsqueeze(-1) * input[:, (curve_parameter_span - self.order).type(torch.LongTensor), :]
            for j in range(1, self.order + 1):
                curves += new_curve_param[:, j].unsqueeze(-1) * input[:, (curve_parameter_span - self.order + j).type(torch.LongTensor), :]
            return curves[:, :, :self._dimension] / curves[:, :, self._dimension].unsqueeze(-1)

    @staticmethod
    def _gen_knot_vector(order: int, num_control_points: int, delta: float = 1e-6) -> np.ndarray:
        """Generate knot vectors automatically.

        Args:
            order (int):
                Order of the curvature.
            num_control_points (int):
                Number of control points.
            delta (float, optional):
                Epsilon to avoid direct 0 value. Defaults to 1e-6.

        Returns:
            np.ndarray: _description_
        """
        # order: degree, n: number of control points; num_control_points+1: number of knots
        num_control_points = order + num_control_points + 1

        # Calculate a uniform interval for middle knots
        num_segments = (num_control_points - 2 * (order + 1) + 1)  # number of segments in the middle
        assert num_segments != 0, 'Number of control points has to be bigger than number of order.'
        spacing = (1.0) / (num_segments)  # spacing between the knots (uniform)

        # First degree+1 knots are "knot_min"
        knot_vector = [(j + 1) * delta for j in range(0, order)]

        # Middle knots
        knot_vector += [mid_knot for mid_knot in np.linspace(delta * (order + 1), 1, num_segments + 1)]

        # Last degree+1 knots are "knot_max"
        knot_vector += [float(1) for j in range(0, order)]

        # Return auto-generated knot vector
        return np.array(knot_vector)



class CurveEvalFunc(torch.autograd.Function):
    """Autograd function for curve evaluation.
    """

    @staticmethod
    def forward(ctx: torch.Tensor, control_points: torch.Tensor, curve_parameter_span: torch.Tensor,
                new_curve_parameter: torch.Tensor, curve_parameter: torch.Tensor,
                num_control_points: int, order: int, _dimension: int, _device: str) -> torch.Tensor:
        """Receive control points, generate curves, and sample points from the curves.

        Args:
            ctx (torch.Tensor):
                Context object.
            control_points (torch.Tensor):
                Control points and the weights of the curvature.
            curve_parameter_span (torch.Tensor):
                Space of the curve parameters.
            new_curve_parameter (torch.Tensor):
                New curve parameters.
            curve_parameter (torch.Tensor):
                Curve parameters.
            num_control_points (int):
                Number of control points.
            order (int):
                Order of the curvature.
            _dimension (int):
                Dimension of the curvature.
            _device (str):
                Device of tensor.

        Returns:
            torch.Tensor:
                Sampled points from the curvature.
        """
        ctx.save_for_backward(control_points)
        ctx.curve_parameter_span = curve_parameter_span
        ctx.new_curve_parameter = new_curve_parameter
        ctx.curve_parameter = curve_parameter
        ctx.num_control_points = num_control_points
        ctx.order = order
        ctx._dimension = _dimension
        ctx._device = _device
        if _device == 'cuda':
            curves = forward(control_points, curve_parameter_span, new_curve_parameter, curve_parameter, num_control_points, order, _dimension)
        else:
            curves = cpp_forward(control_points.cpu(), curve_parameter_span.cpu(), new_curve_parameter.cpu(), curve_parameter.cpu(), num_control_points, order, _dimension)
        ctx.curves = curves
        return curves[:, :, :_dimension] / curves[:, :, _dimension].unsqueeze(-1)

    @staticmethod
    def backward(ctx: torch.Tensor, grad_output: torch.Tensor) -> tuple[Variable, None, None, None, None, None, None, None]:
        """Backward computation of the function.

        Args:
            ctx (torch.Tensor):
                Context objects.
            grad_output (torch.Tensor):
                Gradient outputs

        Returns:
            tuple[Variable, None, None, None, None, None, None, None]:
                The backwards calculation for each of the forward input.
                Everything else is None because they don't need gradients.
        """
        control_points, = ctx.saved_tensors
        curve_parameter_span = ctx.curve_parameter_span
        new_curve_parameter = ctx.new_curve_parameter
        curve_parameter = ctx.curve_parameter
        num_control_points = ctx.num_control_points
        order = ctx.order
        _device = ctx._device
        _dimension = ctx._dimension
        curves = ctx.curves
        grad_cw = torch.zeros((grad_output.size(0), grad_output.size(1), _dimension + 1), dtype=torch.float32)
        if _device == 'cuda':
            grad_cw = grad_cw.cuda()
        grad_cw[:, :, :_dimension] = grad_output
        for d in range(_dimension):
            grad_cw[:, :, _dimension] += grad_output[:, :, d] / curves[:, :, _dimension]
        if _device == 'cuda':
            grad_control_points = backward(grad_cw, control_points, curve_parameter_span, new_curve_parameter, curve_parameter, num_control_points, order, _dimension)
        else:
            grad_control_points = cpp_backward(grad_cw, control_points, curve_parameter_span, new_curve_parameter, curve_parameter, num_control_points, order, _dimension)

        return Variable(grad_control_points[0]), None, None, None, None, None, None, None
